

-------------------------------------------------------------------------------
--! @file       POET_datapath.vhd
--! @author     William Diehl
--! @brief      POET POET_datapath
--! @date       10 July 2016
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity POET_datapath is
	 Generic (G_MAX_LEN : integer:= 28; -- 2^27 - 1 blocks * 128 bits = 2^32 - 1 bytes max length
				 G_DBLK_SIZE : integer:= 128;
				 G_KEY_SIZE : integer:= 128;
				 G_TAG_SIZE : integer:= 128;
				 G_LBS_BYTES: integer:= 4
	 ); 
    Port ( 
			  clk : in STD_LOGIC;
			  rst : in STD_LOGIC;
				--INTERFACE TO INPUT PROCESSOR
			  bdi : in  STD_LOGIC_VECTOR (G_DBLK_SIZE -1 downto 0); --public data input
			  bdi_size : in std_logic_vector(G_LBS_BYTES + 1 - 1 downto 0);
           bdi_valid_bytes : in std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);		  
			  key : in  STD_LOGIC_VECTOR (G_KEY_SIZE - 1 downto 0); --secret data input
			  decrypt: in STD_LOGIC; --high if decrypt
			  decrypt_init : in std_logic;
			  decrypt_reg_out: out std_logic;
			 
			  --bdi_eot: in STD_LOGIC;
			  
			  --INTERFACE TO OUTPUT PROCESSOR
			  bdo : out  STD_LOGIC_VECTOR (G_DBLK_SIZE -1 downto 0);
			  			  
			  --INTERFACE TO CONTROLLER
						
			   load_key : in std_logic;
				S_init : in std_logic;
				block_count_init : in std_logic;
				en_store_bdi: in std_logic;
				M_select : in std_logic_vector(1 downto 0);
				x_sel : in std_logic_vector(1 downto 0);
				bdo_select : in std_logic_vector(1 downto 0);
				right_shift_select : in std_logic_vector(1 downto 0);
				left_shift_select : in std_logic;
				
				block_count_en : in std_logic;
				load_bdi_valid_bytes : in std_logic;
				first_AD_block_init : in std_logic;
				first_AD_block_en : in std_logic;
				first_M_block_init : in std_logic;
				first_M_block_en : in std_logic;
				first_M_block : out std_logic;
				en_last : in std_logic;	
			   tag_valid : out std_logic;
			   En_2iL: in STD_LOGIC;
			   en_X: in STD_LOGIC;
				en_X_decrypt : in std_logic;
				en_tau_alpha_valid: in std_logic;
				x_decrypt_sel : in std_logic;
			   AES_hash_start: in STD_LOGIC;
			  AES_hash_init: in STD_LOGIC;
			  AES_hash_done: out STD_LOGIC;
			  AES_hash_done_init: out STD_LOGIC;
			  AES_hash_ready: out STD_LOGIC;
			  AES_hash_sel : in std_logic;
			  AES_select: in STD_LOGIC_VECTOR(2 downto 0);
			  sel_const: in STD_LOGIC_VECTOR(2 downto 0);
			  keysel: in STD_LOGIC;
			  init_sigma: in STD_LOGIC;
			  en_sigma: in STD_LOGIC;
			  AES_main_start: in STD_LOGIC;
			  AES_main_decrypt: in STD_LOGIC;
			  AES_main_init: in STD_LOGIC;
			  AES_main_done: out STD_LOGIC;
			  AES_main_done_init: out STD_LOGIC;
			  AES_main_ready: out STD_LOGIC;
			  no_M: in std_logic;
			  en_Y: in STD_LOGIC;
			  en_tau: in STD_LOGIC;
			  en_K: in STD_LOGIC;
			  en_L: in STD_LOGIC;
			  en_S: in STD_LOGIC;
			  en_Kf: in STD_LOGIC;
			  en_T_alpha : in STD_LOGIC;
			  en_T : in std_logic;
			  en_T_beta_prime : in std_logic;
			  tag_processing : in STD_LOGIC;
			  init_X_and_Y : in STD_LOGIC
			);
			  
end POET_datapath;

architecture Structural of POET_datapath is

constant length_constant: std_logic_vector(127 - G_MAX_LEN - 4 - 3 downto 0):=(OTHERS=>'0');
constant zeros: std_logic_vector(127 downto 0) := (OTHERS => '0');
signal tau: std_logic_vector (127 downto 0);
signal x_tau, y_tau : std_logic_vector(127 downto 0);
signal Kf: std_logic_vector (127 downto 0);
signal S: std_logic_vector (127 downto 0);
signal x: std_logic_vector (127 downto 0);
signal x_reg: std_logic_vector (127 downto 0);
signal X_decrypt_reg : std_logic_vector(127 downto 0);
signal decrypt_reg : std_logic;

signal exp_tag : std_logic_vector(G_TAG_SIZE - 1 downto 0);
signal hash_din: std_logic_vector (127 downto 0);
signal M: std_logic_vector (127 downto 0);
signal L: std_logic_vector (127 downto 0);
signal reg_2iL: std_logic_vector (127 downto 0);
signal L_GF_2 : std_logic_vector (127 downto 0);
signal key_const: std_logic_vector (127 downto 0);
signal sigma_path, tau_path: std_logic_vector (127 downto 0);
signal encrypt_din: std_logic_vector (127 downto 0);
signal encrypt_key: std_logic_vector (127 downto 0);
signal enc_dout: std_logic_vector (127 downto 0);
signal sigma_next: std_logic_vector (127 downto 0);
signal sigma: std_logic_vector (127 downto 0);
signal sigma_reg: std_logic_vector (127 downto 0);
signal Y: std_logic_vector (127 downto 0);
signal K: std_logic_vector (127 downto 0);
signal input_length_bits_128: std_logic_vector(127 downto 0);
signal input_length: std_logic_vector(127 downto 0);
signal right_shift_input : std_logic_vector(127 downto 0);
signal right_shift_result : std_logic_vector(127 downto 0);
signal right_shift_distance : std_logic_vector(4 downto 0);
signal right_shift_distance_prime : std_logic_vector(4 downto 0);
signal left_shift_input : std_logic_vector(127 downto 0);
signal left_shift_result : std_logic_vector(127 downto 0);
signal left_shift_distance : std_logic_vector(4 downto 0);
signal left_shift_distance_prime : std_logic_vector(4 downto 0);

signal AES_main_next_to_last : std_logic;
signal result_2LHm1_2L : std_logic_vector(127 downto 0);

signal first_AD_block : std_logic;
signal next_block_count: std_logic_vector(G_MAX_LEN - 1 downto 0);
signal block_count : std_logic_vector(G_MAX_LEN - 1 downto 0);

signal block_rem_len : std_logic_vector(G_LBS_BYTES downto 0);
signal bdi_valid_bytes_reg : std_logic_vector(G_DBLK_SIZE/8 - 1 downto 0);
signal trunc_bdi_masking, trunc_bdi_reg_masking, trunc_bdi_reg_rev_masking : std_logic_vector(127 downto 0);

signal Ft : std_logic_vector(127 downto 0);
signal Ft_XOR_M : std_logic_vector(127 downto 0);
signal Ft_XOR_M_XOR_S : std_logic_vector(127 downto 0);
signal Ft_XOR_T_XOR_S : std_logic_vector(127 downto 0);
signal Ft_XOR_Ek : std_logic_vector(127 downto 0);
signal Ft_XOR_S : std_logic_vector(127 downto 0);
signal ft_xor_tau : std_logic_vector(127 downto 0);
signal ek_xor_x_xor_s : std_logic_vector(127 downto 0);

signal Ft_XOR_Ek_XOR_S : std_logic_vector(127 downto 0);
signal Ft_XOR_Ek_XOR_tau : std_logic_vector(127 downto 0);
signal Y_reg: std_logic_vector(127 downto 0);

signal tag : std_logic_vector(127 downto 0);
signal last_block_adjusted : std_logic_vector(127 downto 0);
signal T_beta_adjusted : std_logic_vector(127 downto 0);
signal T_alpha, T_alpha_next, T_beta : std_logic_vector(127 downto 0);
signal T, next_T : std_logic_vector(127 downto 0);
signal trunc_bdi, trunc_bdi_reg, bdi_reg: std_logic_vector(G_DBLK_SIZE - 1 downto 0);
signal trunc_bdi_input, trunc_bdi_reg_input: std_logic_vector(G_DBLK_SIZE - 1 downto 0);
signal bdi_size_reg : std_logic_vector(G_LBS_BYTES + 1 - 1 downto 0);
signal next_S, last : std_logic_vector(127 downto 0);
signal key_reg : std_logic_vector(G_KEY_SIZE - 1 downto 0);
signal x_or_y_reg, x_or_y_reg_xor_s : std_logic_vector(G_DBLK_SIZE - 1 downto 0);
signal tau_alpha : std_logic_vector(127 downto 0);
signal tau_alpha_valid, next_tau_alpha_valid, T_beta_valid: std_logic;
signal next_T_beta_prime, T_beta_prime : std_logic_vector(127 downto 0);
signal tau_alpha_prime : std_logic_vector(127 downto 0);
signal next_bdi_valid_bytes : std_logic_vector(G_DBLK_SIZE/8 - 1 downto 0);

begin

next_block_count <= block_count + 1;
block_rem_len <= bdi_size when (decrypt_reg = '0') else bdi_size_reg;

Blk_Cnt: entity work.Block_Counter (behavioral)
	generic map(G_MAX_LEN => G_MAX_LEN)
	port map (clk => clk,
				 en => block_count_en,
				 init => block_count_init,
				 c_in => next_block_count,
				 c_out => block_count);

next_bdi_valid_bytes <= (OTHERS=>'0') when (block_count_init = '1') else bdi_valid_bytes;

store_valid_bytes: entity work.register_generic(Behavioral) 
				  generic map(width => G_DBLK_SIZE/8 )
				  port map (clk => clk, 
								rst => rst, 
								en => load_bdi_valid_bytes,
								data => next_bdi_valid_bytes, 
								q => bdi_valid_bytes_reg); 

sigma_path <= L_gf_2 xor bdi;
tau_path <= sigma_reg;

input_length <=  zeros when (no_M = '1') else
					  length_constant & block_count & block_rem_len(3 downto 0) & "000" when (block_rem_len(4) = '0') else
					  length_constant & (block_count + 1) & block_rem_len(3 downto 0) & "000";

little_endian: entity work.to_little_endian(dataflow)
			port map (input => input_length,
			          result => input_length_bits_128);

x_tau <= tau when (decrypt_reg = '0') else (tau(127 downto 1) & (tau(0) xor '1'));
y_tau <= (tau(127 downto 1) & (tau(0) xor '1')) when (decrypt_reg = '0') else tau;

Ft_XOR_S <= Ft XOR S;
Ft_XOR_T_XOR_S <= (Ft_XOR_S XOR tau) when (decrypt_reg = '0') else (Ft_XOR_S XOR exp_tag);
Ft_XOR_tau <= Ft xor tau;

with x_sel select
			X <= x_tau when "00",
			Ft_XOR_M_XOR_S when "01",
			Ft_XOR_tau when "10",
			enc_dout when "11",
			x_tau when others;
	  
Y <= y_tau when (init_X_and_Y = '1') else enc_dout;
						 
store_key: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => load_key, 
							data => key, 
							q => key_reg);

store_X: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => En_X, 
							data => X, 
							q => X_reg);
							
store_X_decrypt: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => en_X_decrypt, 
							data => X, 
							q => X_decrypt_reg);

AES_hash: entity work.AES_Four(structure)
			 port map( clk => clk,
						  rst => rst,        
						  din => hash_din,           
						  key => Kf,         
						  dout => Ft,  
						  init => AES_hash_init,      
						  start => AES_hash_start,     
						  ready => AES_hash_ready,     
						  done => AES_hash_done,     
						  done_init => AES_hash_done_init
						  );
						  
Ek_xor_X_xor_S <= enc_dout xor X_reg xor S;

x_or_y_reg <= X_reg when (AES_hash_sel = '0') else Y_reg;
x_or_y_reg_xor_s <= X_reg xor S when (decrypt_reg = '0') else Y_reg;

hash_din <= (X_decrypt_reg xor S) when (x_decrypt_sel = '1') else
				x_or_y_reg_xor_s when (tag_processing = '1') else --decrypt_last_block_reg when (tag_processing = '1') else
				x_or_y_reg;
				
Ft_XOR_M <= Ft XOR M;
Ft_XOR_M_XOR_S <= Ft_XOR_M XOR S; -- S is zero before the last block 

with right_shift_select select
    right_shift_input <=  tau when "00",
								  exp_tag when "01",
								  T when "10", --Ft_XOR_Ek_XOR_tau is buffered in T
	                       T when "11",
			                 T when others;
								  
right_shift_distance_prime <= bdi_size when (decrypt_reg= '0') else bdi_size_reg;
right_shift_distance <= right_shift_distance_prime(4) & (16 - right_shift_distance_prime(3 downto 0)) when (right_shift_select(1) = '1')
                        else right_shift_distance_prime;

r_shift : entity work.variable_right_shift(structural)
					port map(input => right_shift_input,  
					         distance => right_shift_distance(3 downto 0),
								result => right_shift_result);

trunc_bdi_input <= trunc_bdi or right_shift_result when (right_shift_distance(4) = '0') else trunc_bdi;
trunc_bdi_reg_input <= trunc_bdi_reg or right_shift_result when (right_shift_distance(4) = '0') else trunc_bdi_reg;

with M_select select
   M <= trunc_bdi_input when "00",
	     trunc_bdi_reg_input when "01",
	 	  bdi when "10",
	     tau when "11",
		  bdi when others;

with AES_select select
	encrypt_din <= X_reg when "000", 
						input_length_bits_128 when "001",
						key_const when "010",
						sigma_path when "011",
						tau_path when "100",
						Ek_xor_X_xor_S when "101",
						L when "111",
						X_reg when others;

encrypt_key <= K when keysel ='0' else key_reg;
									
main_encrypt: entity work.AES_EncDec(structure)
				  port map(clk => clk,
							  rst => rst,        
							  din => encrypt_din,        
							  key => encrypt_key,        
							  dout => enc_dout, 
							  next_to_last => AES_main_next_to_last,
							  
							  init => AES_main_init,  
							  done_init => AES_main_done_init,
							  decrypt => AES_main_decrypt,
							  start => AES_main_start,     
							  ready => AES_main_ready,     
							  done => AES_main_done);

key_const <= x"0000000000000000000000000000000" & '0' & sel_const;
				  
store_y: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => En_Y, 
							data => Y, 
							q => Y_reg);
					
store_tau: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => En_tau, 
							data => enc_dout, 
							q => tau);
							
store_K: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => En_K, 
							data => enc_dout, 
							q => K);

store_L: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => En_L, 
							data => enc_dout, 
							q => L);
							
next_S <= zeros when (S_init = '1') else enc_dout;
							
store_S: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => En_S, 
							data => next_S, 
							q => S);

store_Kf: entity work.register_generic(Behavioral) 
			  port map (clk => clk, 
							rst => rst, 
							en => En_Kf, 
							data => enc_dout, 
							q => Kf);

store_sigma: entity work.register_generic(Behavioral) 
				  port map (clk => clk, 
								rst => rst, 
								en => En_sigma, 
								data => sigma, 
								q => sigma_reg); 
	
sigma_next <= enc_dout xor sigma_reg;

sigma <= sigma_next when init_sigma ='0' else
			zeros;

Ft_XOR_Ek <= Ft XOR enc_dout;
Ft_XOR_Ek_XOR_tau <= Ft_XOR_Ek XOR tau;
Ft_XOR_Ek_XOR_S <= Ft_XOR_Ek XOR S; -- S is always zero before it is needed

-- Truncate Last Block

	last_block_adjusted <= last and trunc_bdi_masking when (decrypt_reg = '0') else last and trunc_bdi_reg_masking;
	
-- Truncate bdi

	trunc_bdi <= bdi and trunc_bdi_masking;
	
	g1: for i in 0 to 15 generate
		trunc_bdi_masking((15 - i) * 8 + 7 downto (15 - i) * 8) <= bdi_valid_bytes(15-i) & 
		bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) &
		bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i) & bdi_valid_bytes(15-i);
	end generate g1;
	
-- Truncate bdi

	trunc_bdi_reg <= bdi_reg and trunc_bdi_reg_masking;
	
	g3: for i in 0 to 15 generate
		trunc_bdi_reg_masking((15 - i) * 8 + 7 downto (15 - i) * 8) <= bdi_valid_bytes_reg(15-i) & 
		bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) &
		bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i);
	end generate g3;

   g4: for i in 0 to 15 generate
		trunc_bdi_reg_rev_masking( i * 8 + 7 downto i * 8) <= bdi_valid_bytes_reg(15-i) & 
		bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) &
		bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i) & bdi_valid_bytes_reg(15-i);
	end generate g4;
	

	tau_alpha <= tau and (not trunc_bdi_reg_rev_masking) when (left_shift_distance(4) = '0') else (OTHERS=>'0');
	
with bdo_select select
	        bdo <= Ft_XOR_Ek_XOR_S when "00", 
						last_block_adjusted when "01",
						tag when "10", 
						tag when "11", 
						Ft_XOR_Ek_XOR_S when others; 

exp_tag <= bdi;
left_shift_distance_prime <= bdi_size_reg when (decrypt_reg = '1') else bdi_size;
left_shift_distance <= left_shift_distance_prime(4) & (16 - left_shift_distance_prime(3 downto 0)) when (left_shift_select = '1')
                        else left_shift_distance_prime;
left_shift_input <= last when (left_shift_select = '0') else exp_tag;

l_shift : entity work.variable_left_shift(structural)
				   port map(input => left_shift_input,
							   distance => left_shift_distance(3 downto 0),
								result => left_shift_result);

-- exp_tag comes directly from bdi
--Ft_XOR_Ek_XOR_S is buffered in last

store_last: entity work.register_generic(Behavioral) 
				  port map (clk => clk, 
								rst => rst, 
								en => en_last, 
								data => Ft_XOR_Ek_XOR_S,  
								q => last);
								
store_bdi: entity work.register_generic(Behavioral) 
				  port map (clk => clk, 
								rst => rst, 
								en => en_store_bdi, 
								data => bdi,  
								q => bdi_reg);
								
store_bdi_size: entity work.register_generic(Behavioral) 
				  generic map(width => G_LBS_BYTES + 1)
				  port map (clk => clk, 
								rst => rst, 
								en => en_store_bdi, 
								data => bdi_size,  
								q => bdi_size_reg);

next_T <= (OTHERS=>'0') when (block_count_init = '1') else Ft_XOR_Ek_XOR_tau;

store_T: entity work.register_generic(Behavioral) 
				  port map (clk => clk, 
								rst => rst, 
								en => en_T, 
								data => next_T,  
								q => T);
								
store_2iL: entity work.register_generic(Behavioral) 
				  port map (clk => clk, 
								rst => rst, 
								en => En_2iL, 
								data => L_gf_2,  
								q => reg_2iL);

result_2LHm1_2L <= L when (first_AD_block = '0') else reg_2iL;

		 
mult_2: entity work.GF128_mul_2(dataflow)
			port map (input => result_2LHm1_2L,
						 result => L_gf_2);

T_alpha_next <= last when (no_M = '1') else
			       left_shift_result when (left_shift_distance(4) = '0')
					 else (OTHERS=> '0');

store_T_alpha: entity work.register_generic(Behavioral) 
						  port map (clk => clk, 
										rst => rst, 
										en => en_T_alpha, 
										data => T_alpha_next, 
										q => T_alpha);

next_T_beta_prime <= zeros when (block_count_init = '1') else
                     left_shift_result when (left_shift_distance(4) = '0') else
							exp_tag;

store_T_beta_prime: entity work.register_generic(Behavioral) 
						  port map (clk => clk, 
										rst => rst, 
										en => en_T_beta_prime, 
										data => next_T_beta_prime, 
										q => T_beta_prime);

tau_alpha_prime <= tau when (no_M = '1') else
						 left_shift_result when (left_shift_distance(4)='0') else
						 (OTHERS=>'0');

next_tau_alpha_valid <= '1' when (tau_alpha_prime = tau_alpha) else '0';

T_beta <= T and trunc_bdi_reg_masking when (left_shift_distance(4)='0') else T;
T_beta_valid <= '1' when (T_beta = T_beta_prime) else '0';
tag_valid <= tau_alpha_valid and T_beta_valid;
T_beta_adjusted <= right_shift_result when (right_shift_distance(4)='0') else T;
tag <= T_alpha or T_beta_adjusted;						

store_decrypt: entity work.reg1 (behavioral)

	PORT MAP(
		clk => clk,
		init => '0',
		en => decrypt_init,
		d => decrypt,
		q => decrypt_reg);
 
store_tau_alpha_valid: entity work.reg1 (behavioral)

	PORT MAP(
		clk => clk,
		init => '0',
		en => en_tau_alpha_valid,
		d => next_tau_alpha_valid,
		q => tau_alpha_valid);


 decrypt_reg_out <= decrypt_reg;


store_1st_AD_block: entity work.reg1 (behavioral)

	PORT MAP(
		clk => clk,
		init => first_AD_block_init,
		en => first_AD_block_en,
		d => '1',
		q => first_AD_block);
 
store_1st_M_block: entity work.reg1 (behavioral)

	PORT MAP(
		clk => clk,
		init => first_M_block_init,
		en => first_M_block_en,
		d => '1',
		q => first_M_block);

end structural;