

-------------------------------------------------------------------------------
--! @file       variable_right_shift.vhd
--! @author     William Diehl
--! @brief      POET variable_right_shift
--! @date       10 July 2016
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;


entity variable_right_shift is
    Port ( input : in  STD_LOGIC_VECTOR (127 downto 0);
           distance : in  STD_LOGIC_VECTOR (3 downto 0);
           result : out  STD_LOGIC_VECTOR (127 downto 0));
end variable_right_shift;

architecture structural of variable_right_shift is

signal x1a, x1b, x2a, x2b, x3a, x3b, x4a, x4b : std_logic_vector(127 downto 0);

begin

x1a <= input;

x1b <= x"00" & input(127 downto 8);

x2a <= x1a when (distance(0)='0') else x1b; -- variable shift right one byte

x2b <= x"0000" & x2a(127 downto 16);

x3a <= x2a when (distance(1)='0') else x2b; -- variable shift right two bytes

x3b <= x"00000000" & x3a(127 downto 32);

x4a <= x3a when (distance(2)='0') else x3b; -- variable shift right four bytes

x4b <= x"0000000000000000" & x4a(127 downto 64);

result <= x4a when (distance(3)='0') else x4b; -- variable shift right eight bytes

end structural;




